import { FlipnoteMeta, FlipnoteFormat, FlipnoteThumbImageFormat, FlipnoteAudioTrack, FlipnoteSoundEffectTrack, FlipnoteSoundEffectFlags } from './types';
import { BaseParser } from './BaseParser';
/**
 * KWZ section types
 * @internal
 */
export type KwzSectionMagic = 'KFH' | 'KTN' | 'KMC' | 'KMI' | 'KSN' | 'KIC';
/**
 * KWZ section map, tracking their offset and length
 * @internal
 */
export type KwzSectionMap = Map<KwzSectionMagic, {
    ptr: number;
    length: number;
}>;
/**
 * KWZ file metadata, stores information about its playback, author details, etc
 */
export interface KwzMeta extends FlipnoteMeta {
    /**
     * Date representing when the file was created
     */
    creationTimestamp: Date;
}
/**
 * KWZ frame metadata, stores information about each frame, like layer depths sound effect usage
 */
export interface KwzFrameMeta {
    /**
     * Frame flags
     */
    flags: number[];
    /**
     * Frame layer sizes
     */
    layerSize: number[];
    /**
     * Frame author's Flipnote Studio ID
     */
    frameAuthor: string;
    /**
     * Frame layer 3D depths
     */
    layerDepth: number[];
    /**
     * Frame sound
     */
    soundFlags: number;
    /**
     * Whether this frame contains photos taken with the console's camera
     */
    cameraFlag: number;
}
/**
 * KWZ parser options for enabling optimizations and other extra features
 */
export type KwzParserSettings = {
    /**
     * Skip full metadata parsing for quickness
     */
    quickMeta: boolean;
    /**
     * Apply special cases for DSi library notes
     */
    dsiLibraryNote: boolean;
    /**
     * Automatically crop out the border around any frames
     */
    borderCrop: boolean;
    /**
     * Nintendo messed up the initial adpcm state for a bunch of the PPM to KWZ conversions on DSi Library. They are effectively random.
     * By default flipnote.js will try to make a best guess, but you can disable this and provide your own state values
     *
     * This is only enabled if `dsiLibraryNote` is also set to `true`
     */
    guessInitialBgmState: boolean;
    /**
     * Manually provide the initial adpcm step index for the BGM track.
     *
     * This is only enabled if `dsiLibraryNote` is also set to `true`
     */
    initialBgmStepIndex: number | null;
    /**
     * Manually provide the initial adpcm predictor for the BGM track.
     *
     * This is only enabled if `dsiLibraryNote` is also set to `true`
     */
    initialBgmPredictor: number | null;
    /**
     * Manually provide an initial adpcm step index for each sound effect track.
     *
     * This is only enabled if `dsiLibraryNote` is also set to `true`
     */
    initialSeStepIndices: number[] | null;
    /**
     * Manually provide an initial adpcm predictor for each sound effect track.
     *
     * This is only enabled if `dsiLibraryNote` is also set to `true`.
     */
    initialSePredictors: number[] | null;
};
/**
 * Parser class for Flipnote Studio 3D's KWZ animation format. KWZ format docs: https://github.com/Flipnote-Collective/flipnote-studio-3d-docs/wiki/KWZ-Format
 * @group File Parser
 */
export declare class KwzParser extends BaseParser {
    #private;
    /**
     * Default KWZ parser settings
     */
    static defaultSettings: KwzParserSettings;
    /**
     * File format type
     */
    static format: FlipnoteFormat;
    /**
     * Animation frame width
     */
    static width: number;
    /**
     * Animation frame height
     */
    static height: number;
    /**
     * Animation frame aspect ratio
     */
    static aspect: number;
    /**
     * Number of animation frame layers
     */
    static numLayers: number;
    /**
     * Number of colors per layer (aside from transparent)
     */
    static numLayerColors: number;
    /**
     * Audio track base sample rate
     */
    static rawSampleRate: number;
    /**
     * Audio output sample rate
     */
    static sampleRate: number;
    /**
     * Which audio tracks are available in this format
     */
    static audioTracks: FlipnoteAudioTrack[];
    /**
     * Which sound effect tracks are available in this format
     */
    static soundEffectTracks: FlipnoteSoundEffectTrack[];
    /**
     * Global animation frame color palette
     */
    static globalPalette: import("./types").FlipnotePaletteColor[];
    /**
     * Public key used for Flipnote verification, in PEM format
     */
    static publicKey: string;
    static matchBuffer(buffer: ArrayBufferLike): boolean;
    /**
     * File format type, reflects {@link KwzParser.format}
     */
    format: FlipnoteFormat;
    /**
     * Custom object tag
     */
    [Symbol.toStringTag]: string;
    /**
     * Animation frame width, reflects {@link KwzParser.width}
     */
    imageWidth: number;
    /**
     * Animation frame height, reflects {@link KwzParser.height}
     */
    imageHeight: number;
    /**
     * Animation frame aspect ratio, reflects {@link KwzParser.aspect}
     */
    aspect: number;
    /**
     * X offset for the top-left corner of the animation frame
     */
    imageOffsetX: number;
    /**
     * Y offset for the top-left corner of the animation frame
     */
    imageOffsetY: number;
    /**
     * Number of animation frame layers, reflects {@link KwzParser.numLayers}
     */
    numLayers: number;
    /**
     * Number of colors per layer (aside from transparent), reflects {@link KwzParser.numLayerColors}
     */
    numLayerColors: number;
    /**
     * key used for Flipnote verification, in PEM format
     */
    publicKey: string;
    /**
     * @internal
     */
    srcWidth: number;
    /**
     * Which audio tracks are available in this format, reflects {@link KwzParser.audioTracks}
     */
    audioTracks: FlipnoteAudioTrack[];
    /**
     * Which sound effect tracks are available in this format, reflects {@link KwzParser.soundEffectTracks}
     */
    soundEffectTracks: FlipnoteSoundEffectTrack[];
    /**
     * Audio track base sample rate, reflects {@link KwzParser.rawSampleRate}
     */
    rawSampleRate: number;
    /**
     * Audio output sample rate, reflects {@link KwzParser.sampleRate}
     */
    sampleRate: number;
    /**
     * Global animation frame color palette, reflects {@link KwzParser.globalPalette}
     */
    globalPalette: import("./types").FlipnotePaletteColor[];
    /**
     * File metadata, see {@link KwzMeta} for structure.
     * @group Meta
     */
    meta: KwzMeta;
    /**
     * Create a new KWZ file parser instance
     * @param arrayBuffer an ArrayBuffer containing file data
     * @param settings parser settings
     */
    constructor(arrayBuffer: ArrayBufferLike, settings?: Partial<KwzParserSettings>);
    /**
     * Decodes the thumbnail image embedded in the Flipnote. Will return a {@link FlipnoteThumbImage} containing JPEG data.
     *
     * Note: For most purposes, you should probably just decode the thumbnail frame to get a higher resolution image.
     * @group Meta
     */
    getThumbnailImage(): {
        format: FlipnoteThumbImageFormat;
        width: number;
        height: number;
        data: ArrayBuffer;
    };
    /**
     * Get the memory meter level for the Flipnote.
     * This is a value between 0 and 1 indicating how "full" the Flipnote is, based on the size calculation formula inside Flipnote Studio 3D.
     *
     * Values will never be below 0, but can be above 1 if the Flipnote is larger than the size limit - it is technically possible to exceed the size limit by one frame.
     *
     * @group Meta
    */
    getMemoryMeterLevel(): number;
    /**
     * Get the color palette indices for a given frame. RGBA colors for these values can be indexed from {@link KwzParser.globalPalette}
     *
     * Returns an array where:
     *  - index 0 is the paper color index
     *  - index 1 is the layer A color 1 index
     *  - index 2 is the layer A color 2 index
     *  - index 3 is the layer B color 1 index
     *  - index 4 is the layer B color 2 index
     *  - index 5 is the layer C color 1 index
     *  - index 6 is the layer C color 2 index
     * @group Image
    */
    getFramePaletteIndices(frameIndex: number): number[];
    /**
     * Get the RGBA colors for a given frame
     *
     * Returns an array where:
     *  - index 0 is the paper color
     *  - index 1 is the layer A color 1
     *  - index 2 is the layer A color 2
     *  - index 3 is the layer B color 1
     *  - index 4 is the layer B color 2
     *  - index 5 is the layer C color 1
     *  - index 6 is the layer C color 2
     * @group Image
    */
    getFramePalette(frameIndex: number): import("./types").FlipnotePaletteColor[];
    getFrameDiffingFlag(frameIndex: number): number;
    /**
     * Determines if a given frame is a video key frame or not. This returns an array of booleans for each layer, since keyframe encoding is done on a per-layer basis.
     * @param frameIndex
     * @group Image
    */
    getIsKeyFrame(frameIndex: number): boolean[];
    /**
     * Get the 3D depths for each layer in a given frame.
     * @param frameIndex
     * @group Image
    */
    getFrameLayerDepths(frameIndex: number): number[];
    /**
     * Get the FSID for a given frame's original author.
     * @param frameIndex
     * @group Meta
    */
    getFrameAuthor(frameIndex: number): string;
    /**
     * Get the camera flags for a given frame
     * @group Image
     * @returns Array of booleans, indicating whether each layer uses a photo or not
    */
    getFrameCameraFlags(frameIndex: number): boolean[];
    /**
     * Get the layer draw order for a given frame
     * @group Image
    */
    getFrameLayerOrder(frameIndex: number): number[];
    /**
     * Decode a frame, returning the raw pixel buffers for each layer
     * @group Image
    */
    decodeFrame(frameIndex: number, diffingFlag?: number, isPrevFrame?: boolean): [Uint8Array, Uint8Array, Uint8Array];
    decodeFrameSoundFlags(frameIndex: number): boolean[];
    /**
     * Get the sound effect flags for every frame in the Flipnote
     * @group Audio
    */
    decodeSoundFlags(): boolean[][];
    /**
     * Get the sound effect usage flags for every frame
     * @group Audio
     */
    getSoundEffectFlags(): FlipnoteSoundEffectFlags[];
    /**
     * Get the sound effect usage for a given frame
     * @param frameIndex
     * @group Audio
     */
    getFrameSoundEffectFlags(frameIndex: number): FlipnoteSoundEffectFlags;
    /**
     * Get the duration of a given track in seconds
     * @returns number
     * @group Audio
     */
    getAudioTrackDuration(trackId: FlipnoteAudioTrack): number;
    /**
     * Get the raw compressed audio data for a given track
     * @returns Byte array
     * @group Audio
    */
    getAudioTrackRaw(trackId: FlipnoteAudioTrack): Uint8Array;
    decodeAdpcm(src: Uint8Array, dst: Int16Array, predictor?: number, stepIndex?: number): number;
    /**
     * Get the decoded audio data for a given track, using the track's native samplerate
     * @returns Signed 16-bit PCM audio
     * @group Audio
    */
    decodeAudioTrack(trackId: FlipnoteAudioTrack): Int16Array;
    /**
     * Get the decoded audio data for a given track, using the specified samplerate
     * @returns Signed 16-bit PCM audio
     * @group Audio
    */
    getAudioTrackPcm(trackId: FlipnoteAudioTrack, dstFreq?: number): Int16Array;
    pcmAudioMix(src: Int16Array, dst: Int16Array, dstOffset?: number): void;
    /**
     * Get the full mixed audio for the Flipnote, using the specified samplerate
     * @returns Signed 16-bit PCM audio
     * @group Audio
    */
    getAudioMasterPcm(dstFreq?: number): Int16Array;
    /**
     * Get the body of the Flipnote - the data that is digested for the signature
     * @group Verification
     */
    getBody(): Uint8Array;
    /**
     * Get the Flipnote's signature data
     * @group Verification
     */
    getSignature(): Uint8Array;
    /**
     * Verify whether this Flipnote's signature is valid
     * @group Verification
     */
    verify(): Promise<boolean>;
}
