import { FlipnoteMeta, FlipnoteFormat, FlipnoteThumbImageFormat, FlipnoteAudioTrack, FlipnoteSoundEffectTrack, FlipnoteSoundEffectFlags } from './types';
import { BaseParser } from './BaseParser';
/**
 * PPM file metadata, stores information about its playback, author details, etc
 */
export interface PpmMeta extends FlipnoteMeta {
    /**
     * In-app frame playback speed when the BGM audio track was recorded
     */
    bgmSpeed: number;
}
/**
 * PPM parser options for enabling optimizations and other extra features.
 * None are currently implemented
 */
export type PpmParserSettings = {};
/**
 * Parser class for (DSiWare) Flipnote Studio's PPM animation format. Format docs: https://github.com/Flipnote-Collective/flipnote-studio-docs/wiki/PPM-format
 * @group File Parser
 */
export declare class PpmParser extends BaseParser {
    #private;
    /**
     * Default PPM parser settings.
     */
    static defaultSettings: PpmParserSettings;
    /**
     * File format type.
     */
    static format: FlipnoteFormat;
    /**
     * Animation frame width.
     */
    static width: number;
    /**
     * Animation frame height.
     */
    static height: number;
    /**
     * Animation frame aspect ratio.
     */
    static aspect: number;
    /**
     * Number of animation frame layers.
     */
    static numLayers: number;
    /**
     * Number of colors per layer (aside from transparent).
     */
    static numLayerColors: number;
    /**
     * Audio track base sample rate.
     */
    static rawSampleRate: number;
    /**
     * Nintendo DSi audio output rate.
     */
    static sampleRate: number;
    /**
     * Which audio tracks are available in this format.
     */
    static audioTracks: FlipnoteAudioTrack[];
    /**
     * Which sound effect tracks are available in this format.
     */
    static soundEffectTracks: FlipnoteSoundEffectTrack[];
    /**
     * Global animation frame color palette.
     */
    static globalPalette: import("./types").FlipnotePaletteColor[];
    /**
     * Public key used for Flipnote verification, in PEM format.
     */
    static publicKey: string;
    static matchBuffer(buffer: ArrayBufferLike): boolean;
    /**
     * File format type, reflects {@link PpmParser.format}.
     * @group Meta
     */
    format: FlipnoteFormat;
    /**
     * Custom object tag.
     * @group Utility
     */
    [Symbol.toStringTag]: string;
    /**
     * Animation frame width, reflects {@link PpmParser.width}.
     * @group Image
     */
    imageWidth: number;
    /**
     * Animation frame height, reflects {@link PpmParser.height}.
     * @group Image
     */
    imageHeight: number;
    /**
     * Animation frame aspect ratio, reflects {@link PpmParser.aspect}.
     * @group Image
     */
    aspect: number;
    /**
     * X offset for the top-left corner of the animation frame.
     * @group Image
     */
    imageOffsetX: number;
    /**
     * Y offset for the top-left corner of the animation frame.
     * @group Image
     */
    imageOffsetY: number;
    /**
     * Number of animation frame layers, reflects {@link PpmParser.numLayers}.
     * @group Image
     */
    numLayers: number;
    /**
     * Number of colors per layer (aside from transparent), reflects {@link PpmParser.numLayerColors}.
     * @group Image
     */
    numLayerColors: number;
    /**
     * Key used for Flipnote verification, in PEM format.
     * @group Verification
     */
    publicKey: string;
    /**
     * @internal
     * @group Image
     */
    srcWidth: number;
    /**
     * Which audio tracks are available in this format, reflects {@link PpmParser.audioTracks}.
     * @group Audio
     */
    audioTracks: FlipnoteAudioTrack[];
    /**
     * Which sound effect tracks are available in this format, reflects {@link PpmParser.soundEffectTracks}.
     * @group Audio
     */
    soundEffectTracks: FlipnoteSoundEffectTrack[];
    /**
     * Audio track base sample rate, reflects {@link PpmParser.rawSampleRate}.
     * @group Audio
     */
    rawSampleRate: number;
    /**
     * Audio output sample rate, reflects {@link PpmParser.sampleRate}.
     * @group Audio
     */
    sampleRate: number;
    /**
     * Global animation frame color palette, reflects {@link PpmParser.globalPalette}.
     * @group Image
     */
    globalPalette: import("./types").FlipnotePaletteColor[];
    /**
     * File metadata, see {@link PpmMeta} for structure.
     * @group Meta
     */
    meta: PpmMeta;
    /**
     * File format version; always the same as far as we know.
     * @group Meta
     */
    version: number;
    /**
     * Create a new PPM file parser instance
     * @param arrayBuffer an ArrayBuffer containing file data
     * @param settings parser settings (none currently implemented)
     */
    constructor(arrayBuffer: ArrayBufferLike, settings?: Partial<PpmParserSettings>);
    /**
     * Decodes the thumbnail image embedded in the Flipnote. Will return a {@link FlipnoteThumbImage} containing raw RGBA data.
     *
     * Note: For most purposes, you should probably just decode the thumbnail frame to get a higher resolution image.
     * @group Meta
     */
    getThumbnailImage(): {
        format: FlipnoteThumbImageFormat;
        width: number;
        height: number;
        data: ArrayBufferLike;
    };
    /**
     * Get the memory meter level for the Flipnote.
     * This is a value between 0 and 1 indicating how "full" the Flipnote is, based on the size limit of Flipnote Studio.
     *
     * Values will never be below 0, but can be above 1 if the Flipnote is larger than the size limit - it is technically possible to exceed the size limit by one frame.
     *
     * @group Meta
    */
    getMemoryMeterLevel(): number;
    /**
     * Decode a frame, returning the raw pixel buffers for each layer
     * @group Image
    */
    decodeFrame(frameIndex: number): [Uint8Array, Uint8Array];
    /**
     * Get the color palette indices for a given frame. RGBA colors for these values can be indexed from {@link PpmParser.globalPalette}
    
     * Returns an array where:
     *  - index 0 is the paper color index
     *  - index 1 is the layer 1 color index
     *  - index 2 is the layer 2 color index
     * @group Image
    */
    getFramePaletteIndices(frameIndex: number): number[];
    /**
     * Get the RGBA colors for a given frame
     *
     * Returns an array where:
     *  - index 0 is the paper color
     *  - index 1 is the layer 1 color
     *  - index 2 is the layer 2 color
     * @group Image
     */
    getFramePalette(frameIndex: number): import("./types").FlipnotePaletteColor[];
    /**
     * Determines if a given frame is a video key frame or not. This returns an array of booleans for each layer, since in the KWZ format, keyframe encoding is done on a per-layer basis.
     * @param frameIndex
     * @group Image
    */
    getIsKeyFrame(frameIndex: number): boolean[];
    /**
     * Get the 3D depths for each layer in a given frame. The PPM format doesn't actually store this information, so `0` is returned for both layers. This method is only here for consistency with KWZ.
     * @param frameIndex
     * @group Image
    */
    getFrameLayerDepths(frameIndex: number): number[];
    /**
     * Get the FSID for a given frame's original author. The PPM format doesn't actually store this information, so the current author FSID is returned. This method is only here for consistency with KWZ.
     * @param frameIndex
     * @group Meta
    */
    getFrameAuthor(frameIndex: number): string;
    /**
     * Get the camera flags for a given frame. The PPM format doesn't actually store this information so `false` will be returned for both layers. This method is only here for consistency with KWZ.
     * @group Image
     * @returns Array of booleans, indicating whether each layer uses a photo or not
    */
    getFrameCameraFlags(frameIndex: number): boolean[];
    /**
     * Get the layer draw order for a given frame
     * @group Image
     * @returns Array of layer indices, in the order they should be drawn
    */
    getFrameLayerOrder(frameIndex: number): number[];
    /**
     * Get the sound effect flags for every frame in the Flipnote
     * @group Audio
    */
    decodeSoundFlags(): boolean[][];
    /**
     * Get the sound effect usage flags for every frame
     * @group Audio
     */
    getSoundEffectFlags(): FlipnoteSoundEffectFlags[];
    /**
     * Get the sound effect usage flags for a given frame
     * @group Audio
     */
    getFrameSoundEffectFlags(frameIndex: number): FlipnoteSoundEffectFlags;
    /**
    * Get the duration of a given track in seconds
    * @returns number
    * @group Audio
    */
    getAudioTrackDuration(trackId: FlipnoteAudioTrack): number;
    /**
     * Get the raw compressed audio data for a given track
     * @returns byte array
     * @group Audio
    */
    getAudioTrackRaw(trackId: FlipnoteAudioTrack): Uint8Array;
    /**
     * Get the decoded audio data for a given track, using the track's native samplerate
     * @returns Signed 16-bit PCM audio
     * @group Audio
    */
    decodeAudioTrack(trackId: FlipnoteAudioTrack): Int16Array;
    /**
     * Get the decoded audio data for a given track, using the specified samplerate
     * @returns Signed 16-bit PCM audio
     * @group Audio
    */
    getAudioTrackPcm(trackId: FlipnoteAudioTrack, dstFreq?: number): Int16Array;
    /**
     * Get the full mixed audio for the Flipnote, using the specified samplerate
     * @returns Signed 16-bit PCM audio
     * @group Audio
    */
    getAudioMasterPcm(dstFreq?: number): Int16Array;
    /**
     * Get the body of the Flipnote - the data that is digested for the signature
     * @group Verification
     */
    getBody(): Uint8Array;
    /**
    * Get the Flipnote's signature data
    * @group Verification
    */
    getSignature(): Uint8Array;
    /**
     * Verify whether this Flipnote's signature is valid
     * @group Verification
     */
    verify(): Promise<boolean>;
}
