import type { Flipnote, FlipnoteFormat, FlipnoteMeta, FlipnoteParserSettings } from '../parsers';
import { PlayerEvent, PlayerEventMap } from './PlayerEvent';
import { UniversalCanvas } from '../renderers';
import { WebAudioPlayer } from '../webaudio';
type PlayerLayerVisibility = Record<number, boolean>;
/**
 * Flipnote Player API (exported as `flipnote.Player`) - provides a {@link https://developer.mozilla.org/en-US/docs/Web/API/HTMLMediaElement | MediaElement}-like interface for loading Flipnotes and playing them.
 * This is intended for cases where you want to implement your own player UI, if you just want a pre-built player with some nice UI controls, check out the [Web Components](/web-components/) page instead!
 *
 * ### Create a new player
 *
 * You'll need an element in your page's HTML to act as a wrapper for the player:
 *
 * ```html
 *  <div id="player-wrapper"></div>
 * ```
 *
 * Then you can create a new `Player` instance by passing a CSS selector that matches the wrapper element, plus the desired width and height.
 *
 * ```js
 *  const player = new flipnote.Player('#player-wrapper', 320, 240);
 * ```
 *
 * ### Load a Flipnote
 *
 * Load a Flipnote from a valid {@link FlipnoteSource}:
 *
 * ```js
 * player.load('./path/to/flipnote.ppm');
 * ```
 *
 * ### Listen to events
 *
 * Use the {@link on} method to register event listeners:
 *
 * ```js
 *  player.on('play', function() {
 *    // do something when the Flipnote starts playing...
 *  });
 * ```
 */
export declare class Player {
    #private;
    /**
     * Flipnote renderer.
     * @group Display Control
     */
    renderer: UniversalCanvas;
    /**
     * Root element
     */
    el: Element;
    /**
     * Currently loaded Flipnote
     */
    note: Flipnote;
    /**
     * Flipnote parser settings
     */
    parserSettings: FlipnoteParserSettings;
    /**
     * Format of the currently loaded Flipnote
     */
    noteFormat: FlipnoteFormat;
    /**
     * Metadata for the currently loaded Flipnote
     */
    meta: FlipnoteMeta;
    /**
     * Animation duration, in seconds
     */
    duration: number;
    /**
     * Animation layer visibility
     */
    layerVisibility: PlayerLayerVisibility;
    /**
     * Automatically begin playback after a Flipnote is loaded
     */
    autoplay: boolean;
    /**
     * Array of events supported by this player
     */
    supportedEvents: PlayerEvent[];
    /**
     * Audio player
     * @internal
     */
    audio: WebAudioPlayer;
    /**
     * @internal
     */
    _src: any;
    /**
     * @internal
     */
    _loop: boolean;
    /**
     * @internal
     */
    _volume: number;
    /**
     * @internal
     */
    _muted: boolean;
    /**
     * @internal
     */
    _frame: number;
    /**
     * @internal
     */
    _hasEnded: boolean;
    /**
     * @internal
     */
    isNoteLoaded: boolean;
    /**
     * @internal
     */
    events: PlayerEventMap;
    /**
     * @internal
     */
    playbackStartTime: number;
    /**
     * @internal
     */
    playbackTime: number;
    /**
     * @internal
     */
    playbackLoopId: number;
    /**
     * @internal
     */
    showThumbnail: boolean;
    /**
     * @internal
     */
    hasPlaybackStarted: boolean;
    /**
     * @internal
     */
    isPlaying: boolean;
    /**
     * @internal
     */
    wasPlaying: boolean;
    /**
     * @internal
     */
    isSeeking: boolean;
    /**
     * Create a new Player instance
     *
     * @param parent - Element to mount the rendering canvas to
     * @param width - Canvas width (pixels)
     * @param height - Canvas height (pixels)
     *
     * The ratio between `width` and `height` should be 3:4 for best results
     */
    constructor(parent: string | Element, width: number, height: number, parserSettings?: FlipnoteParserSettings);
    /**
     * The currently loaded Flipnote source, if there is one
     * @group Lifecycle
     * @deprecated
     */
    get src(): any;
    set src(source: any);
    /**
     * Indicates whether playback is currently paused
     * @group Playback Control
     */
    get paused(): boolean;
    set paused(isPaused: boolean);
    /**
     * Current animation frame index.
     * @group Playback Control
     */
    get currentFrame(): number;
    set currentFrame(frameIndex: number);
    /**
     * Current animation playback position, in seconds.
     * @group Playback Control
     */
    get currentTime(): number;
    set currentTime(value: number);
    /**
     * Current animation playback progress, as a percentage out of 100.
     * @group Playback Control
     */
    get progress(): number;
    set progress(value: number);
    /**
     * Audio volume, range `0` to `1`.
     * @group Audio Control
     */
    get volume(): number;
    set volume(value: number);
    /**
     * Audio mute state.
     * @group Audio Control
     */
    get muted(): boolean;
    set muted(value: boolean);
    /**
     * Indicates whether playback should loop once the end is reached
     * @group Playback Control
     */
    get loop(): boolean;
    set loop(value: boolean);
    /**
     * Animation frame rate, measured in frames per second.
     * @group Playback Control
     */
    get framerate(): number;
    /**
     * Animation frame count.
     * @group Frame Control
     */
    get frameCount(): number;
    /**
     * Animation frame speed.
     * @group Frame Control
     */
    get frameSpeed(): number;
    /**
     * Implementation of the `HTMLMediaElement` {@link https://developer.mozilla.org/en-US/docs/Web/API/HTMLMediaElement/buffered | buffered } property.
     * @group HTMLVideoElement compatibility
     */
    get buffered(): TimeRanges;
    /**
     * Implementation of the `HTMLMediaElement` {@link https://developer.mozilla.org/en-US/docs/Web/API/HTMLMediaElement/seekable | seekable} property
     * @group HTMLVideoElement compatibility
     */
    get seekable(): TimeRanges;
    /**
     * Implementation of the `HTMLMediaElement` {@link https://developer.mozilla.org/en-US/docs/Web/API/HTMLMediaElement/currentSrc | currentSrc} property
     * @group HTMLVideoElement compatibility
     */
    get currentSrc(): any;
    /**
     * Implementation of the `HTMLVideoElement` {@link https://developer.mozilla.org/en-US/docs/Web/API/HTMLVideoElement/videoWidth | videoWidth} property
     * @group HTMLVideoElement compatibility
     */
    get videoWidth(): number;
    /**
     * Implementation of the `HTMLVideoElement` {@link https://developer.mozilla.org/en-US/docs/Web/API/HTMLVideoElement/videoHeight | videoHeight} property
     * @group HTMLVideoElement compatibility
     */
    get videoHeight(): number;
    /**
     * Open a Flipnote from a source
     * @group Lifecycle
     */
    load(source: any): Promise<void>;
    /**
     * Reload the current Flipnote.
     * @group Lifecycle
     */
    reload(): Promise<void>;
    /**
     * Reload the current Flipnote, applying new parser settings.
     * @group Lifecycle
     */
    updateSettings(settings: FlipnoteParserSettings): Promise<void>;
    /**
     * Close the currently loaded Flipnote
     * @group Lifecycle
     */
    closeNote(): void;
    /**
     * Open a Flipnote into the player
     * @group Lifecycle
     */
    openNote(note: Flipnote): void;
    /**
     * Playback animation loop
     * @internal
     * @group Playback Control
     */
    playbackLoop: (timestamp: DOMHighResTimeStamp) => void;
    /**
     * Set the current playback time
     * @group Playback Control
     */
    setCurrentTime(value: number): void;
    /**
     * Get the current playback time
     * @group Playback Control
     */
    getCurrentTime(): number;
    /**
     * Get the current time as a counter string, like `"0:00 / 1:00"`
     * @group Playback Control
     */
    getTimeCounter(): string;
    /**
     * Get the current frame index as a counter string, like `"001 / 999"`
     * @group Playback Control
     */
    getFrameCounter(): string;
    /**
     * Set the current playback progress as a percentage (`0` to `100`)
     * @group Playback Control
     */
    setProgress(value: number): void;
    /**
     * Get the current playback progress as a percentage (0 to 100)
     * @group Playback Control
     */
    getProgress(): number;
    /**
     * Begin animation playback starting at the current position
     * @group Playback Control
     */
    play(): Promise<void>;
    /**
     * Pause animation playback at the current position
     * @group Playback Control
     */
    pause(): void;
    /**
     * Resumes animation playback if paused, otherwise pauses
     * @group Playback Control
     */
    togglePlay(): void;
    /**
     * Determines if playback is currently paused
     * @group Playback Control
     */
    getPaused(): boolean;
    /**
     * Get the duration of the Flipnote in seconds
     * @group Playback Control
     */
    getDuration(): number;
    /**
     * Determines if playback is looped
     * @group Playback Control
     */
    getLoop(): boolean;
    /**
     * Set the playback loop
     * @group Playback Control
     */
    setLoop(loop: boolean): void;
    /**
     * Switch the playback loop between on and off
     * @group Playback Control
     */
    toggleLoop(): void;
    /**
     * Jump to a given animation frame
     * @group Frame Control
     */
    setCurrentFrame(newFrameValue: number): void;
    /**
     * Jump to the next animation frame
     * If the animation loops, and is currently on its last frame, it will wrap to the first frame
     * @group Frame Control
     */
    nextFrame(): void;
    /**
     * Jump to the next animation frame
     * If the animation loops, and is currently on its first frame, it will wrap to the last frame
     * @group Frame Control
     */
    prevFrame(): void;
    /**
     * Jump to the last animation frame
     * @group Frame Control
     */
    lastFrame(): void;
    /**
     * Jump to the first animation frame
     * @group Frame Control
     */
    firstFrame(): void;
    /**
     * Jump to the thumbnail frame
     * @group Frame Control
     */
    thumbnailFrame(): void;
    /**
     * Begins a seek operation
     * @group Playback Control
     */
    startSeek(): void;
    /**
     * Seek the playback progress to a different position
     * @param position - animation playback position, range `0` to `1`
     * @group Playback Control
     */
    seek(position: number): void;
    /**
     * Ends a seek operation
     * @group Playback Control
     */
    endSeek(): void;
    /**
     * Draws the specified animation frame to the canvas. Note that this doesn't update the playback time or anything, it simply decodes a given frame and displays it.
     * @param frameIndex
     * @group Display Control
     */
    drawFrame(frameIndex: number): void;
    /**
     * Forces the current animation frame to be redrawn
     * @group Display Control
     */
    forceUpdate(): void;
    /**
     * Resize the playback canvas to a new size
     * @param width - new canvas width (pixels)
     * @param height - new canvas height (pixels)
     *
     * The ratio between `width` and `height` should be 3:4 for best results
     *
     * @group Display Control
     */
    resize(width: number, height: number): void;
    /**
     * Sets whether an animation layer should be visible throughout the entire animation
     * @param layer - layer index, starting at 1
     * @param value - `true` for visible, `false` for invisible
     *
     * @group Display Control
     */
    setLayerVisibility(layer: number, value: boolean): void;
    /**
     * Returns the visibility state for a given layer
     * @param layer - layer index, starting at 1
     *
     * @group Display Control
     */
    getLayerVisibility(layer: number): boolean;
    /**
     * Toggles whether an animation layer should be visible throughout the entire animation
     *
     * @group Display Control
     */
    toggleLayerVisibility(layerIndex: number): void;
    /**
     * Toggle audio Sudomemo equalizer filter.
     * @group Audio Control
     */
    toggleAudioEq(): void;
    /**
     * Turn audio Sudomemo equalizer filter on or off.
     * @group Audio Control
     */
    setAudioEq(state: boolean): void;
    /**
     * Turn the audio off
     * @group Audio Control
     */
    mute(): void;
    /**
     * Turn the audio on
     * @group Audio Control
     */
    unmute(): void;
    /**
     * Turn the audio on or off
     * @group Audio Control
     */
    setMuted(isMute: boolean): void;
    /**
     * Get the audio mute state
     * @group Audio Control
     */
    getMuted(): boolean;
    /**
     * Switch the audio between muted and unmuted
     * @group Audio Control
     */
    toggleMuted(): void;
    /**
     * Set the audio volume
     * @group Audio Control
     */
    setVolume(volume: number): void;
    /**
     * Get the current audio volume
     * @group Audio Control
     */
    getVolume(): number;
    /**
     * Implementation of the `HTMLVideoElement` {@link https://developer.mozilla.org/en-US/docs/Web/API/HTMLVideoElement/seekToNextFrame | seekToNextFrame} method
     * @group HTMLVideoElement compatibility
     */
    seekToNextFrame(): void;
    /**
     * Implementation of the `HTMLMediaElement` {@link https://developer.mozilla.org/en-US/docs/Web/API/HTMLMediaElement/fastSeek | fastSeek} method
     * @group HTMLVideoElement compatibility
     */
    fastSeek(time: number): void;
    /**
     * Implementation of the `HTMLVideoElement` {@link https://developer.mozilla.org/en-US/docs/Web/API/HTMLVideoElement/getVideoPlaybackQuality | getVideoPlaybackQuality } method
     * @group HTMLVideoElement compatibility
     */
    canPlayType(mediaType: string): "" | "probably" | "maybe";
    /**
     * Implementation of the `HTMLVideoElement` [https://developer.mozilla.org/en-US/docs/Web/API/HTMLVideoElement/getVideoPlaybackQuality](getVideoPlaybackQuality) method
     * @group HTMLVideoElement compatibility
     */
    getVideoPlaybackQuality(): VideoPlaybackQuality;
    /**
     * Implementation of the `HTMLVideoElement` [https://developer.mozilla.org/en-US/docs/Web/API/HTMLVideoElement/requestPictureInPicture](requestPictureInPicture) method. Not currently working, only a stub.
     * @group HTMLVideoElement compatibility
     */
    requestPictureInPicture(): void;
    /**
     * Implementation of the `HTMLVideoElement` [https://developer.mozilla.org/en-US/docs/Web/API/HTMLVideoElement/captureStream](captureStream) method. Not currently working, only a stub.
     * @group HTMLVideoElement compatibility
     */
    captureStream(): void;
    /**
     * Fired when animation playback begins or is resumed
     * @event
     */
    onplay: () => void;
    /**
     * Fired when animation playback is paused
     * @event
     */
    onpause: () => void;
    /**
     * Fired when the Flipnote has loaded enough to begin animation play
     * @group HTMLVideoElement compatibility
     */
    oncanplay: () => void;
    /**
     * Fired when the Flipnote has loaded enough to play successfully
     * @group HTMLVideoElement compatibility
     */
    oncanplaythrough: () => void;
    /**
     * Fired when a seek operation begins
     * @event
     */
    onseeking: () => void;
    /**
     * Fired when a seek operation completes
     * @event
     */
    onseeked: () => void;
    /**
     * Fired when the animation duration has changed
     * @group HTMLVideoElement compatibility
     */
    ondurationchange: () => void;
    /**
     * Fired when playback has looped after reaching the end
     * @event
     */
    onloop: () => void;
    /**
     * Fired when playback has ended
     * @event
     */
    onended: () => void;
    /**
     * Fired when the player audio volume or muted state has changed
     * @event
     */
    onvolumechange: (volume: number) => void;
    /**
     * Fired when playback progress has changed
     * @event
     */
    onprogress: (progress: number) => void;
    /**
     * Fired when the playback time has changed
     * @event
     */
    ontimeupdate: (time: number) => void;
    /**
     * Fired when the current frame index has changed
     * @event
     */
    onframeupdate: (frameIndex: number) => void;
    /**
     * Fired when {@link nextFrame} has been called
     * @event
     */
    onframenext: () => void;
    /**
     * Fired when {@link prevFrame} has been called
     * @event
     */
    onframeprev: () => void;
    /**
     * Fired when {@link firstFrame} has been called
     * @event
     */
    onframefirst: () => void;
    /**
     * Fired when {@link lastFrame} has been called
     * @event
     */
    onframelast: () => void;
    /**
     * Fired when a Flipnote is ready for playback
     * @event
     */
    onready: () => void;
    /**
     * Fired when a Flipnote has finished loading
     * @event
     */
    onload: () => void;
    /**
     * Fired when a Flipnote has begun loading
     * @event
     */
    onloadstart: () => void;
    /**
     * Fired when the Flipnote data has been loaded; implementation of the `HTMLMediaElement` [https://developer.mozilla.org/en-US/docs/Web/API/HTMLMediaElement/loadeddata_event](loadeddata) event.
     * @group HTMLVideoElement compatibility
     */
    onloadeddata: () => void;
    /**
     * Fired when the Flipnote metadata has been loaded; implementation of the `HTMLMediaElement` [https://developer.mozilla.org/en-US/docs/Web/API/HTMLMediaElement/loadedmetadata_event](loadedmetadata) event.
     * @group HTMLVideoElement compatibility
     */
    onloadedmetadata: () => void;
    /**
     * Fired when the media has become empty; implementation of the `HTMLMediaElement` [https://developer.mozilla.org/en-US/docs/Web/API/HTMLMediaElement/emptied_event](emptied) event.
     * @group HTMLVideoElement compatibility
     */
    onemptied: () => void;
    /**
     * Fired after the Flipnote has been closed with {@link close}
     * @event
     */
    onclose: () => void;
    /**
     * Fired after a loading, parsing or playback error occurs
     * @event
     */
    onerror: (err?: Error) => void;
    /**
     * Fired just before the player has been destroyed, after calling {@link destroy}
     * @event
     */
    ondestroy: () => void;
    /**
     * Add an event callback
     * @group Event API
     */
    on(eventType: PlayerEvent | PlayerEvent[], listener: Function): void;
    /**
     * Remove an event callback
     * @group Event API
     */
    off(eventType: PlayerEvent | PlayerEvent[], callback: Function): void;
    /**
     * Emit an event - mostly used internally
     * @group Event API
     */
    emit(eventType: PlayerEvent, ...args: any): void;
    /**
     * Remove all registered event callbacks
     * @group Event API
     */
    clearEvents(): void;
    /**
     * Destroy a Player instance
     * @group Lifecycle
     */
    destroy(): Promise<void>;
    /**
     * @internal
     */
    assertNoteLoaded(): void;
}
export {};
