import { type CanvasInterface, CanvasStereoscopicMode } from './CanvasInterface';
import { type Flipnote } from '../parsers';
/**
 * Settings for {@link WebGlCanvas}.
 */
export interface WebglCanvasOptions {
    /**
     * Function to be called if the context is lost.
     */
    onlost: () => void;
    /**
     * Function to be called if the context is restored.
     */
    onrestored: () => void;
    /**
     * Use DPI scaling.
     * @default true
     */
    useDpi: boolean;
}
/**
 * Flipnote renderer for the {@link https://developer.mozilla.org/en-US/docs/Web/API/WebGL_API WebGL} API.
 *
 * Only available in browser contexts.
 */
export declare class WebglCanvas implements CanvasInterface {
    #private;
    static defaultOptions: WebglCanvasOptions;
    static isSupported(): boolean;
    /**
     * Flipnote parser object for the Flipnote being played.
     */
    note: Flipnote;
    /**
     * Canvas HTML element being used as a rendering surface.
     */
    canvas: HTMLCanvasElement;
    /**
     * Rendering context - see {@link https://developer.mozilla.org/en-US/docs/Web/API/WebGLRenderingContext}
     */
    gl: WebGLRenderingContext;
    /**
     * View width (CSS pixels)
     */
    width: number;
    /**
     * View height (CSS pixels)
     */
    height: number;
    /**
     * Flipnote width (pixels).
     */
    srcWidth: number;
    /**
     * Flipnote height (pixels).
     */
    srcHeight: number;
    /**
     * Backing canvas width (real pixels)
     * Note that this factors in device pixel ratio, so it may not reflect the size of the canvas in CSS pixels
     */
    dstWidth: number;
    /**
     * Backing canvas height (real pixels)
     * Note that this factors in device pixel ratio, so it may not reflect the size of the canvas in CSS pixels
     */
    dstHeight: number;
    /**
     *
     */
    frameIndex: number;
    /**
     *
     */
    supportedStereoscopeModes: CanvasStereoscopicMode[];
    /**
     *
     */
    stereoscopeMode: CanvasStereoscopicMode;
    /**
     *
     */
    stereoscopeStrength: number;
    /**
     * Creates a new WebGlCanvas instance
     * @param el - Canvas HTML element to use as a rendering surface
     * @param width - Canvas width in CSS pixels
     * @param height - Canvas height in CSS pixels
     *
     * The ratio between `width` and `height` should be 3:4 for best results
     */
    constructor(parent: Element, width?: number, height?: number, options?: Partial<WebglCanvasOptions>);
    /**
     * Resize the canvas surface
     * @param width - New canvas width, in CSS pixels
     * @param height - New canvas height, in CSS pixels
     *
     * The ratio between `width` and `height` should be 3:4 for best results
     */
    setCanvasSize(width: number, height: number): void;
    /**
     * Sets the note to use for this player
     */
    setNote(note: Flipnote): void;
    /**
     * Clear the canvas
     * @param color optional RGBA color to use as a background color
     */
    clear(color?: [number, number, number, number]): void;
    /**
     * Draw a frame from the currently loaded Flipnote
     * @param frameIndex
     */
    drawFrame(frameIndex: number): void;
    requestStereoScopeMode(mode: CanvasStereoscopicMode): void;
    forceUpdate(): void;
    /**
     * Returns true if the webGL context has returned an error
     */
    isErrorState(): boolean;
    /**
     * Get the contents of the canvas as a data URL.
     *
     * @param type image mime type (`image/jpeg`, `image/png`, etc)
     * @param quality image quality where supported, between 0 and 1
     */
    getDataUrl(type?: string, quality?: any): string;
    /**
     * Get the contents of the canvas as a blob.
     *
     * @param type image mime type (`image/jpeg`, `image/png`, etc)
     * @param quality image quality where supported, between 0 and 1
     */
    getBlob(type?: string, quality?: any): Promise<Blob>;
    /**
     * Frees any resources used by this canvas instance
     */
    destroy(): void;
}
